/* cftest.c
   CompactFlash benchmark tool

Copyright 2005-2007 Damian Yerrick <p8@pineight.com>

This work is provided 'as-is', without any express or implied
warranty.  In no event will the authors be held liable for any
damages arising from the use of this work.

Permission is granted to anyone to use this work for any purpose,
including commercial applications, and to alter it and redistribute
it freely, subject to the following restrictions:

1. The origin of this work must not be misrepresented; you must
   not claim that you wrote the original work. If you use
   this work in a product, an acknowledgment in the product
   documentation would be appreciated but is not required.
2. Altered source versions must be plainly marked as such, and must
   not be misrepresented as being the original work.
3. This notice may not be removed or altered from any source
   distribution.

"Source" is the preferred form of a work for making changes to it.

*/

#include <stdio.h>
#include <string.h>

#ifdef ARM9
#include <nds.h>
#define VBlankIntrWait swiWaitForVBlank
#define TIMER_PER_SEC 33514000
#else
#include <gba_video.h>
#include <gba_interrupt.h>
#include <gba_input.h>
#include <gba_dma.h>
#include <gba_systemcalls.h>
#define TIMER_PER_SEC 16777216
#endif
#include <fat.h>
#include "agbtty.h"
#include "talkback.h"

#include <sys/dir.h>
#include <sys/stat.h>
#include <sys/syslimits.h>

#define BIG_FILE_SIZE (256 * 1024)

static void press_A(void)
{
  while(!(REG_KEYINPUT & KEY_A))
    VBlankIntrWait();
  while( (REG_KEYINPUT & KEY_A))
    VBlankIntrWait();
}

static void halt() __attribute__ ((noreturn));

#ifdef ARM9
/* My ghetto IPC system */
volatile P8A7Talkback tb = {
  .needPowerOff = 0,
  .needNote = 0
};

static void SoftReset(int unused) {
  // Because of the lack of clarity of the license of Lick's
  // rebootlib, I'll need to use the turn-DS-off code here
  tb.needPowerOff = 1;
  DC_FlushAll();
}
#endif

static void halt(void)
{
  VBlankIntrWait();
  BG_PALETTE[0] = RGB5(0, 0, 0);
  VBlankIntrWait();
  SoftReset(0);
  while (1) {
    VBlankIntrWait();
  }
}

void ls(void)
{
  char filename[NAME_MAX + 1];
  struct stat st;

  DIR_ITER *dp = diropen(".");
  
  while (!dirnext(dp, filename, &st)) {
    agbtty_puts(filename);
	
	if (st.st_mode & S_IFDIR) {
	  agbtty_putc('/');
	}
    agbtty_putc('\n');
  }
  dirclose(dp);
}


#if 0
/* find_big_file() ********************
   Finds a file in the current directory at least |minimum| bytes
   in size.  If it finds one, returns 1 and copies its short name
   to |filename|.  If not, returns 0.
*/
int find_big_file(char *out_name, u32 minimum) {
  char filename[NAME_MAX + 1];
  struct stat st;

  DIR_ITER *dp = diropen(".");
  
  while (!dirnext(dp, filename, &st)) {
    if(!(st.st_mode & S_IFDIR))
    {
      u32 size = st.st_size;
      if(size >= minimum)
      {
        strcpy(out_name, filename);
        dirclose(dp);
        return 1;
      }
    }
  }
  dirclose(dp);
  return 0;
}
#endif

static void agbtty_putn(int n) {
  char txt[16];
  siprintf(txt, "%d", n);
  agbtty_puts(txt);
}

static unsigned int cycles2micro(unsigned int read_time) {
  return read_time * 1000000ULL / TIMER_PER_SEC;
}


#define TIMER ((volatile u32 *)(0x04000100))
#define TMCNT(x) ((x) << 16)
// libgba and libnds differ on naming
#ifndef TIMER_ENABLE
#define TIMER_ENABLE (1 << 7)
#endif
#ifndef TIMER_CASCADE
#define TIMER_CASCADE (1 << 2)
#endif

void startTimer(void) {
  TIMER[2] = 0;
  TIMER[3] = 0;
  TIMER[3] = 0 | TMCNT(TIMER_CASCADE | TIMER_ENABLE);
  TIMER[2] = 0 | TMCNT(TIMER_ENABLE);
}

u32 readTimer(void) {
  unsigned int lo = TIMER[2] & 0xFFFF;
  unsigned int hi = TIMER[3] & 0xFFFF;

  // If the timer wrapped between reading lo and hi, reread once.
  unsigned int lo2 = TIMER[2] & 0xFFFF;
  if (lo2 < lo) {
    hi = TIMER[3] & 0xFFFF;
    lo = lo2;
  }
  return lo | (hi << 16);
}

static unsigned int ayn_seed;
static unsigned int ayn(u32 max)
{
  ayn_seed = ayn_seed * 1103515245 + 12345;
  return ((ayn_seed >> 16) * max) >> 16;
}

enum {
  BENCHMARK_RANDOM = 1,
  BENCHMARK_QUIET = 2
};

static unsigned char *dst;

static int benchmark_read(const char *filename, u32 read_size, int flags, u32 *out)
{
  FILE *fp = fopen(filename, "r");
  if (!fp) {
    if (!(flags & BENCHMARK_QUIET)) {
      agbtty_puts("\fThe file\n");
      agbtty_puts(filename);
      agbtty_puts("\nwas not found.");
      press_A();
    }
    return 0;
  }

  u32 max_seek = BIG_FILE_SIZE;
  unsigned int read_times[4] = {0, 0, 0, 0};
  unsigned int time_left = 80;
  unsigned int total_time = 0, total_max_time = 0, n_blocks_read = 0;
  unsigned int sector = 0;

  if(max_seek < read_size * 4)
  {
    if (!(flags & BENCHMARK_QUIET)) {
      agbtty_puts("\fneed ... bigger ... file ...\n");
      agbtty_puts(filename);
      agbtty_puts("\ndoesn't cut it.");
      press_A();
    }
    fclose(fp);
    return 0;
  }
  max_seek /= read_size;
  
  if (!(flags & BENCHMARK_QUIET)) {
    agbtty_puts("\fHow fast can your card read\n");
    agbtty_putn(read_size);
    agbtty_puts("-byte chunks from\n");
    agbtty_puts(filename);
    agbtty_puts(flags & BENCHMARK_RANDOM
                ? "\nin random order?\n"
                : "\nin sequential order?\n");
    agbtty_puts("(smaller numbers are faster)");
  }
  
  do {
    read_times[3] = read_times[2];
    read_times[2] = read_times[1];
    read_times[1] = read_times[0];

    if (flags & BENCHMARK_RANDOM) {
      sector = ayn(max_seek) * read_size;
    }

    /* time each read */
    startTimer();
    if (flags & BENCHMARK_RANDOM) {
      fseek(fp, sector, SEEK_SET);
    }
	if (fread(dst, 1, read_size, fp)) {
      unsigned int max_time, avg_time;
      u32 value = *(u32 *)dst;
      u32 expectedValue = (sector >> 9) * 123456789;

	  read_times[0] = readTimer();
      ++n_blocks_read;
      total_time += read_times[0];
      avg_time = total_time / n_blocks_read;
      sector += read_size;

      /* take the longest of the last 4 read times to avoid flicker */
      max_time = read_times[0];
      if (max_time < read_times[1])
        max_time = read_times[1];
      if (max_time < read_times[2])
        max_time = read_times[2];
      if (max_time < read_times[3])
        max_time = read_times[3];
      total_max_time += max_time;

      /* if it's not reading the card, signal this */
      if (value != expectedValue) {
        total_time = 0;
        n_blocks_read = 0;
      }

      /* report this time */
      if (!(flags & BENCHMARK_QUIET)) {
        char dst[128];
        agbtty_gotoxy(0, 5);
        siprintf(dst,
                 "max:%7u cycles (%5u \265s)\n"
                 "avg:%7u cycles (%5u \265s)\n", 
	             max_time,
			     cycles2micro(max_time),
                 avg_time,
                 cycles2micro(avg_time));
        agbtty_puts(dst);
        if (value != expectedValue) {
          agbtty_puts("Not actually reading the card!");
        }
      }
    } else {
      rewind(fp);
      sector = 0;
    }

    /* slow it down a bit */
    VBlankIntrWait();
    if (!(flags & BENCHMARK_QUIET)) {
      VBlankIntrWait();
      VBlankIntrWait();
    }
    time_left--;
  } while (time_left > 0);

  fclose(fp);
  
  
  if (n_blocks_read > 1) {
    if (out) {
      out[0] = total_max_time / n_blocks_read;
      out[1] = total_time / n_blocks_read;
    }
    return 1;
  } else {
    return 0;
  }
}

void initHW(void) {
  // set up ISR for VBlankIntrWait()
  irqInit();
  irqEnable(IRQ_VBLANK);
#ifdef ARM9  // defined on DS, not on GBA
  powerON(POWER_ALL_2D);
  vramSetMainBanks(VRAM_A_MAIN_BG, 0x82, 
                   VRAM_C_SUB_BG, VRAM_D_SUB_SPRITE);
  videoSetMode(MODE_0_2D
               | DISPLAY_BG0_ACTIVE);
  IPC->soundData = (void *)&tb;
#else
  SetMode(MODE_0 | BG0_ON);
#endif
  dst = malloc(16384);
}

#define N_TESTS 6
#define N_MENU_ITEMS (N_TESTS + 3)

struct TestDef {
  unsigned int blockSize;
  unsigned int flags;
};

const struct TestDef testDefs[N_TESTS] = {
  {  512, 0},
  { 4096, 0},
  {16384, 0},
  {  512, BENCHMARK_RANDOM},
  { 4096, BENCHMARK_RANDOM},
  {16384, BENCHMARK_RANDOM}
};

void getTestName(char *dst, unsigned int defNo) {
  const char *seqRandom = testDefs[defNo].flags & BENCHMARK_RANDOM
                          ? "Random"
                          : "Sequential";
  unsigned int KiB = testDefs[defNo].blockSize >= 1024;
  const char *unitLabel = KiB ? "KiB" : "B";
  unsigned int reportedSize = KiB
                              ? testDefs[defNo].blockSize / 1024
                              : testDefs[defNo].blockSize;
  
  siprintf(dst, "%s (%u %s)", seqRandom, reportedSize, unitLabel);
}

void runAllTests(const char *big_filename) {
  agbtty_puts("\fAll tests: average times\n\n");

  for (int i = 0; i < N_TESTS; ++i) {
    char line[32];
    getTestName(line, i);
    agbtty_puts(line);
    agbtty_putc('\n');
    agbtty_putc('\n');
  }
  for (int i = 0; i < N_TESTS; ++i) {
    u32 times[2];
    int got = benchmark_read(big_filename,
                             testDefs[i].blockSize,
                             testDefs[i].flags
                             | BENCHMARK_QUIET,
                             times);
    agbtty_gotoxy(0, 3 + i * 2);
    if (got) {
      char line[32];
      siprintf(line,
               " max:%6u \265s; avg:%6u \265s",
               cycles2micro(times[0]),
               cycles2micro(times[1]));
      agbtty_puts(line);
    } else {
      agbtty_puts("Error");
    }
  }
  
  press_A();
}


//---------------------------------------------------------------------------------
// Program entry point
//---------------------------------------------------------------------------------
int main(void)
{
  unsigned int menu_y = 0;
  const char *big_filename = "test_sectors.bin";
  
  initHW();

  // set up AGBTTY
  agbtty_init();

  // set up palette
  BG_PALETTE[0] = RGB5(31,31,23);
  BG_PALETTE[1] = RGB5( 0, 0, 0);
  SPRITE_PALETTE[1] = RGB5(16,16, 0);
  
  if (!fatInitDefault())
  {
    agbtty_puts("libfat failed to recognize the\n"
                "adapter. Do you need DLDI?\n");
    press_A();
    halt();
  }

  while(menu_y != N_MENU_ITEMS - 1)
  {
    int selected = 0;
    u32 last_joy = 0x3FF;

    agbtty_puts("\fSpeed tester for libfat\n"
                " 2005-2007 Damian Yerrick\n\n");
    // list all tests
    for (int i = 0; i < N_TESTS; ++i) {
      char line[32];
      getTestName(line, i);
      agbtty_putc(' ');
      agbtty_putc(' ');
      agbtty_puts(line);
      agbtty_putc('\n');
    }
    agbtty_puts("  Run all tests\n"
                "  List root directory\n"
                "  Quit\n"
                "\n(Up and Down select; A starts)");
    agbtty_gotoxy(0, 3 + menu_y);
    agbtty_putc('>');
    while(!selected)
    {
      u32 j = ~REG_KEYINPUT;
      u32 jnew = j & ~last_joy;
      u32 last_y = menu_y;

      last_joy = j;
      if(jnew & KEY_DOWN)
      {
        menu_y++;
        if(menu_y >= N_MENU_ITEMS)
          menu_y = 0;
      }
      if(jnew & KEY_UP)
      {
        if(menu_y == 0)
          menu_y = N_MENU_ITEMS - 1;
        else
          menu_y--;
      }
      if(jnew & KEY_B)
      {
        menu_y = N_MENU_ITEMS - 1;
      }

      if(jnew & KEY_A)
      {
        selected = 1;
      }
      
      // move cursor if needed
      if (last_y != menu_y) {
        agbtty_gotoxy(0, 3 + last_y);
        agbtty_putc(' ');
        agbtty_gotoxy(0, 3 + menu_y);
        agbtty_putc('>');
      }
      VBlankIntrWait();
    }

    switch(menu_y) {
    case N_TESTS:
      runAllTests(big_filename);
      break;
    case N_TESTS + 1:
      startTimer();
      agbtty_putc('\f');
      ls();
      press_A();
      break;
    default:
      if (menu_y < N_TESTS) {
        benchmark_read(big_filename,
                       testDefs[menu_y].blockSize,
                       testDefs[menu_y].flags,
                       NULL);
      }
      break;
    }
  }
  REG_IME = 0;
  halt();
}


